"use strict";
/*======================================================================================
    Space Y: An Asteroids Clone

    Copyright 2022 by TheDummyProgrammer (https://www.thedummyprogrammer.com)

    This file is part of Space Y.

    Space Y is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Space Y is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Space Y.  If not, see <http://www.gnu.org/licenses/>.
======================================================================================*/
class Point {
    constructor(x, y) {
        this._x = x;
        this._y = y;
    }
    get x() {
        return this._x;
    }
    set x(value) {
        this._x = value;
    }
    get y() {
        return this._y;
    }
    set y(value) {
        this._y = value;
    }
    toString() {
        return `Point: x:${this._x}, y:${this._y}`;
    }
}
class Size {
    constructor(width, height) {
        this._width = 0;
        this._height = 0;
        this._width = width;
        this._height = height;
    }
    get width() {
        return this._width;
    }
    set width(value) {
        this._width = value;
    }
    get height() {
        return this._height;
    }
    set height(value) {
        this._height = value;
    }
    toString() {
        return `Size: width:${this._width}, height:${this._height}`;
    }
}
class Rectangle {
    constructor(location, size) {
        this._x = 0;
        this._y = 0;
        this._width = 0;
        this._height = 0;
        this._x = location.x;
        this._y = location.y;
        this._width = size.width;
        this._height = size.height;
    }
    get x() {
        return this._x;
    }
    set x(value) {
        this._x = value;
    }
    get y() {
        return this._y;
    }
    set y(value) {
        this._y = value;
    }
    get width() {
        return this._width;
    }
    set width(value) {
        this._width = value;
    }
    get height() {
        return this._height;
    }
    set height(value) {
        this._height = value;
    }
    get topLeft() {
        return new Point(this.x, this.y);
    }
    get topRight() {
        return new Point(this.x + this.width, this.y);
    }
    get bottomRight() {
        return new Point(this.x + this.width, this.y + this.height);
    }
    get bottomLeft() {
        return new Point(this.x, this.y + this.height);
    }
    offset(x, y) {
        this.x += x;
        this.y += y;
    }
    clone() {
        return new Rectangle(new Point(this._x, this._y), new Size(this._width, this._height));
    }
    contains(pt) {
        return (pt.x >= this._x && pt.x <= (this._x + this._width) && pt.y >= this._y && pt.y <= (this._y + this._height));
    }
    intersectsWith(rect) {
        return (this.contains(rect.topLeft) || this.contains(rect.topRight) || this.contains(rect.bottomRight) || this.contains(rect.bottomLeft));
    }
    toString() {
        return `Rectangle: x:${this._x}, y:${this._y}, width:${this._width}, height:${this._height}`;
    }
}
class AppParameters {
}
AppParameters.gameAreaUpdateDelayMS = 20;
AppParameters.accelerationIncrement = 0.3;
AppParameters.accelerationDecrement = 0.1;
AppParameters.rotationIncrement = 4;
AppParameters.rotationDecrement = 4;
AppParameters.maxSpeed = 12;
AppParameters.bulletTimeToLive = 60;
AppParameters.bulletInitialSpeed = 14;
AppParameters.bulletInitialRotation = 90;
AppParameters.boundingBoxDecrementPercent = 20;
AppParameters.shipInitialRotation = 90;
AppParameters.asteroidRotationIncrement = 1;
AppParameters.maxUfoShips = 3;
AppParameters.maxAsteroids = 5;
AppParameters.maxPlayerLives = 3;
AppParameters.scoreAsteroidBig = 20;
AppParameters.scoreAsteroidMedium = 50;
AppParameters.scoreAsteroidSmall = 100;
AppParameters.scoreUfoShip = 200;
AppParameters.debugDrawBoundingBoxLocal = false;
AppParameters.debugDrawBoundingBox = false;
class AxisTranslation {
    constructor(origin) {
        this._origin = origin;
    }
    get origin() {
        return this._origin;
    }
    set origin(value) {
        this._origin = value;
    }
    toExtPoint(pt) {
        return new Point(this._origin.x + pt.x, this._origin.y - pt.y);
    }
    toIntPoint(pt) {
        return new Point(pt.x - this._origin.x, this._origin.y - pt.y);
    }
}
class Utility {
    static degreeToRadians(degree) {
        return degree * (Math.PI / 180);
    }
    static radiansToDegree(radians) {
        return radians * (180 / Math.PI);
    }
    static transformPoint(point, rotation, zoom) {
        let radRotation = Utility.degreeToRadians(rotation - 90);
        let newX = (point.x * Math.cos(radRotation) - point.y * Math.sin(radRotation)) * zoom;
        let newY = (point.x * Math.sin(radRotation) + point.y * Math.cos(radRotation)) * zoom;
        return new Point(newX, newY);
    }
    static getRandomInteger(min, max) {
        return Math.floor(min + Math.random() * (max - min + 1));
    }
}
class DrawHelper {
    constructor(canvasContext, origin) {
        this._ctx = canvasContext;
        this._axisTrn = new AxisTranslation(origin);
    }
    get drawContext() {
        return this._ctx;
    }
    set drawContext(value) {
        this._ctx = value;
    }
    get axisTranslation() {
        return this._axisTrn;
    }
    set axisTranslation(value) {
        this._axisTrn = value;
    }
    beginPath() {
        this._ctx.beginPath();
    }
    moveTo(pt) {
        let extPt = this._axisTrn.toExtPoint(pt);
        this._ctx.moveTo(extPt.x, extPt.y);
    }
    lineTo(pt) {
        let extPt = this._axisTrn.toExtPoint(pt);
        this._ctx.lineTo(extPt.x, extPt.y);
    }
    rect(rect) {
        let topLeft = this._axisTrn.toExtPoint(rect.topLeft);
        this._ctx.rect(topLeft.x, topLeft.y, rect.width, rect.height);
    }
    get strokeStyle() {
        return this._ctx.strokeStyle;
    }
    set strokeStyle(value) {
        this._ctx.strokeStyle = value;
    }
    stroke() {
        this._ctx.stroke();
    }
}
class Vector {
    constructor(magnitude, direction) {
        this._magnitude = 0;
        this._direction = 0;
        this._magnitude = magnitude;
        this._direction = direction;
    }
    get magnitude() {
        return this._magnitude;
    }
    set magnitude(value) {
        if (value > 0)
            this._magnitude = value;
        else
            this._magnitude = 0;
    }
    get direction() {
        return this._direction;
    }
    get directionRad() {
        return Utility.degreeToRadians(this._direction);
    }
    set direction(value) {
        if (value >= 0 && value <= 360)
            this._direction = value;
        else if (value > 360)
            this._direction = value % 360;
        else if (value < 0)
            this._direction = (360 - (value % 360));
    }
    toPoint() {
        return new Point(this._magnitude * Math.cos(this.directionRad), this._magnitude * Math.sin(this.directionRad));
    }
    add(vectorToAdd) {
        let thisVectorPoint = this.toPoint();
        let otherVectorPoint = vectorToAdd.toPoint();
        let newX = thisVectorPoint.x + otherVectorPoint.x;
        let newY = thisVectorPoint.y + otherVectorPoint.y;
        let newMagnitude = Math.sqrt((newX * newX) + (newY * newY));
        let newDirection = 0;
        if (newX === 0 && newY > 0) {
            newDirection = 90;
        }
        else if (newX === 0 && newY < 0) {
            newDirection = 270;
        }
        else if (newX === 0 && newY === 0) {
            newDirection = 0;
        }
        else if (newX > 0 && newY >= 0) {
            newDirection = Math.atan(newY / newX);
        }
        else if (newX > 0 && newY < 0) {
            newDirection = Math.atan(newY / newX) + (Math.PI * 2);
        }
        else if (newX < 0) {
            newDirection = Math.atan(newY / newX) + Math.PI;
        }
        this._magnitude = newMagnitude;
        this._direction = Utility.radiansToDegree(newDirection);
    }
    toString() {
        return `Vector (Magnitude: ${this._magnitude}, Direction: ${this._direction})`;
    }
}
class GameObject {
    constructor(mainDrawHelper, position) {
        this._position = new Point(0, 0);
        this._rotation = AppParameters.shipInitialRotation;
        this._vectorMotionChange = null;
        this._vectorMotion = null;
        this._zoom = 1;
        this._destroyed = false;
        this._lines = new Set();
        this._polygons = new Set();
        this._mainDrawHelper = mainDrawHelper;
        this._position = position;
    }
    get position() {
        return this._position;
    }
    set position(value) {
        this._position = value;
    }
    get rotation() {
        return this._rotation;
    }
    set rotation(value) {
        if (value >= 0 && value <= 360)
            this._rotation = value;
        else if (value > 360)
            this._rotation = value % 360;
        else if (value < 0)
            this._rotation = (360 - (value % 360));
    }
    get vectorMotionChange() {
        return this._vectorMotionChange;
    }
    set vectorMotionChange(value) {
        this._vectorMotionChange = value;
    }
    get vectorMotion() {
        return this._vectorMotion;
    }
    set vectorMotion(value) {
        this._vectorMotion = value;
    }
    get zoom() {
        return this._zoom;
    }
    set zoom(value) {
        this._zoom = value;
    }
    get destroyed() {
        return this._destroyed;
    }
    set destroyed(value) {
        this._destroyed = value;
    }
    buildObject() {
    }
    init() {
        this.buildObject();
    }
    updatePosition() {
    }
    getLocalBoundingBox() {
        let upLeftPoint = null;
        let bottomRightPoint = null;
        this._lines.forEach((value, key, s) => {
            let pointsTransformed = value.map((v, i, a) => {
                return Utility.transformPoint(v, this._rotation, this._zoom);
            });
            pointsTransformed.forEach((v, i, a) => {
                if (upLeftPoint === null) {
                    upLeftPoint = new Point(v.x, v.y);
                    bottomRightPoint = new Point(v.x, v.y);
                }
                else {
                    if (v.x < upLeftPoint.x)
                        upLeftPoint.x = v.x;
                    if (v.y > upLeftPoint.y)
                        upLeftPoint.y = v.y;
                    if (v.x > bottomRightPoint.x)
                        bottomRightPoint.x = v.x;
                    if (v.y < bottomRightPoint.y)
                        bottomRightPoint.y = v.y;
                }
            });
        });
        this._polygons.forEach((value, key, s) => {
            let pointsTransformed = value.map((v, i, a) => {
                return Utility.transformPoint(v, this._rotation, this._zoom);
            });
            pointsTransformed.forEach((v, i, a) => {
                if (upLeftPoint === null) {
                    upLeftPoint = new Point(v.x, v.y);
                    bottomRightPoint = new Point(v.x, v.y);
                }
                else {
                    if (v.x < upLeftPoint.x)
                        upLeftPoint.x = v.x;
                    if (v.y > upLeftPoint.y)
                        upLeftPoint.y = v.y;
                    if (v.x > bottomRightPoint.x)
                        bottomRightPoint.x = v.x;
                    if (v.y < bottomRightPoint.y)
                        bottomRightPoint.y = v.y;
                }
            });
        });
        // Reduce the bounding box
        upLeftPoint.x *= (100 - AppParameters.boundingBoxDecrementPercent) / 100;
        upLeftPoint.y *= (100 - AppParameters.boundingBoxDecrementPercent) / 100;
        bottomRightPoint.x *= (100 - AppParameters.boundingBoxDecrementPercent) / 100;
        bottomRightPoint.y *= (100 - AppParameters.boundingBoxDecrementPercent) / 100;
        return new Rectangle(upLeftPoint, new Size(Math.abs(bottomRightPoint.x - upLeftPoint.x), Math.abs(bottomRightPoint.y - upLeftPoint.y)));
    }
    getBoundingBox() {
        let localBBox = this.getLocalBoundingBox();
        let extPoint = this._mainDrawHelper.axisTranslation.toExtPoint(this._position);
        let localAxisTranslation = new AxisTranslation(extPoint);
        return new Rectangle(localAxisTranslation.toExtPoint(localBBox.topLeft), new Size(localBBox.width, localBBox.height));
    }
    drawLine(points) {
        let pointsTransformed = points.map((v, i, a) => {
            return Utility.transformPoint(v, this._rotation, this._zoom);
        });
        let extPoint = this._mainDrawHelper.axisTranslation.toExtPoint(this._position);
        this._localDrawHelper = new DrawHelper(this._mainDrawHelper.drawContext, extPoint);
        this._localDrawHelper.beginPath();
        for (let idxPointFrom = 0; idxPointFrom < pointsTransformed.length; idxPointFrom++) {
            let idxPointTo = idxPointFrom + 1;
            if (idxPointFrom === (points.length - 1))
                break;
            this._localDrawHelper.moveTo(pointsTransformed[idxPointFrom]);
            this._localDrawHelper.lineTo(pointsTransformed[idxPointTo]);
        }
        this._localDrawHelper.strokeStyle = 'white';
        this._localDrawHelper.stroke();
    }
    drawPolygon(points) {
        let pointsTransformed = points.map((v, i, a) => {
            return Utility.transformPoint(v, this._rotation, this._zoom);
        });
        let extPoint = this._mainDrawHelper.axisTranslation.toExtPoint(this._position);
        this._localDrawHelper = new DrawHelper(this._mainDrawHelper.drawContext, extPoint);
        this._localDrawHelper.beginPath();
        for (let idxPointFrom = 0; idxPointFrom < pointsTransformed.length; idxPointFrom++) {
            let idxPointTo = idxPointFrom + 1;
            if (idxPointFrom === (points.length - 1))
                idxPointTo = 0;
            this._localDrawHelper.moveTo(pointsTransformed[idxPointFrom]);
            this._localDrawHelper.lineTo(pointsTransformed[idxPointTo]);
        }
        this._localDrawHelper.strokeStyle = 'white';
        this._localDrawHelper.stroke();
    }
    draw() {
        this._lines.forEach((value, key, s) => {
            this.drawLine(value);
        });
        this._polygons.forEach((value, key, s) => {
            this.drawPolygon(value);
        });
        if (AppParameters.debugDrawBoundingBoxLocal) {
            let localBBox = this.getLocalBoundingBox();
            this._localDrawHelper.beginPath();
            this._localDrawHelper.strokeStyle = 'red';
            this._localDrawHelper.rect(localBBox);
            this._localDrawHelper.stroke();
        }
    }
}
class Asteroid extends GameObject {
    constructor(mainDrawHelper, position) {
        super(mainDrawHelper, position);
        this._rotationIncrement = AppParameters.asteroidRotationIncrement;
        this._type = 0;
        this._size = 1;
        this._vectorMotion = new Vector(Utility.getRandomInteger(1, 3), Utility.getRandomInteger(0, 360));
        this._rotation = Utility.getRandomInteger(0, 360);
        this._rotationIncrement = Utility.getRandomInteger(0, 1);
        this._size = Utility.getRandomInteger(1, 3);
        this._zoom = this._size;
        this._type = Utility.getRandomInteger(0, 4);
    }
    get type() {
        return this._type;
    }
    set type(value) {
        this._type = value;
    }
    get size() {
        return this._size;
    }
    set size(value) {
        this._zoom = value;
        this._size = value;
    }
    buildObject() {
        let points;
        if (this._type === 0) {
            points = [new Point(-2, 12), new Point(6, 12), new Point(10, 6), new Point(10, 2), new Point(4, -10),
                new Point(-4, -10), new Point(0, 0), new Point(-8, -10), new Point(-12, -4), new Point(-6, -2),
                new Point(-10, 2)];
        }
        else if (this._type === 1) {
            points = [new Point(-4, 10), new Point(4, 10), new Point(10, 2), new Point(6, -4), new Point(8, -10),
                new Point(-2, -10), new Point(-8, -4), new Point(-8, 6)];
        }
        else if (this._type === 2) {
            points = [new Point(-4, 10), new Point(10, 6), new Point(14, 6), new Point(12, -8), new Point(-2, -12),
                new Point(-10, -4), new Point(-4, 2)];
        }
        else if (this._type === 3) {
            points = [new Point(6, 10), new Point(-6, 10), new Point(-10, 6), new Point(-10, 0), new Point(-10, -6),
                new Point(-6, -10), new Point(6, -10), new Point(6, 0), new Point(-2, 0), new Point(-2, -4),
                new Point(2, -4), new Point(2, -6), new Point(-4, -6), new Point(-6, -4), new Point(-6, 4),
                new Point(-4, 6), new Point(6, 6)];
        }
        else {
            points = [new Point(-4, 8), new Point(8, 8), new Point(8, 4), new Point(0, 4), new Point(0, -0),
                new Point(6, 0), new Point(6, -2), new Point(0, -2), new Point(0, -10), new Point(-4, -10)];
        }
        this._polygons.add(points);
    }
    updatePosition() {
        let deltaX = this._vectorMotion.magnitude * Math.cos(this._vectorMotion.directionRad);
        let deltaY = this._vectorMotion.magnitude * Math.sin(this._vectorMotion.directionRad);
        this._position.x += deltaX;
        this._position.y += deltaY;
        if (this._rotationIncrement === 1)
            this._rotation++;
        else
            this._rotation--;
    }
}
class Bullet extends GameObject {
    constructor(mainDrawHelper, position) {
        super(mainDrawHelper, position);
        this._timeToLive = AppParameters.bulletTimeToLive;
        this._zoom = 1;
        this._vectorMotion = new Vector(AppParameters.bulletInitialSpeed, AppParameters.bulletInitialRotation);
        this._rotation = AppParameters.bulletInitialRotation;
    }
    buildObject() {
        let points = [new Point(0, 0), new Point(0, 3)];
        this._lines.add(points);
    }
    updatePosition() {
        let deltaX = this._vectorMotion.magnitude * Math.cos(this._vectorMotion.directionRad);
        let deltaY = this._vectorMotion.magnitude * Math.sin(this._vectorMotion.directionRad);
        this._position.x += deltaX;
        this._position.y += deltaY;
        this._timeToLive--;
        if (this._timeToLive <= 0)
            this._destroyed = true;
    }
}
class UfoBullet extends Bullet {
}
class UfoShip extends GameObject {
    constructor(mainDrawHelper, position) {
        super(mainDrawHelper, position);
        this._vectorMotion = new Vector(Utility.getRandomInteger(1, 3), Utility.getRandomInteger(0, 360));
        this._rotation = 90;
        this._zoom = 2;
    }
    buildObject() {
        let points = [new Point(3, 6), new Point(6, 3), new Point(12, 0), new Point(6, -3), new Point(-6, -3),
            new Point(-12, 0), new Point(-6, 3), new Point(-3, 6)];
        this._polygons.add(points);
        let line1 = [new Point(-12, 0), new Point(12, 0)];
        this._lines.add(line1);
        let line2 = [new Point(-6, 3), new Point(6, 3)];
        this._lines.add(line2);
        this._gunPosition = new Point(0, -3);
    }
    updatePosition() {
        let deltaX = this._vectorMotion.magnitude * Math.cos(this._vectorMotion.directionRad);
        let deltaY = this._vectorMotion.magnitude * Math.sin(this._vectorMotion.directionRad);
        this._position.x += deltaX;
        this._position.y += deltaY;
    }
    fire(userShipPosition) {
        let bulletPos = Utility.transformPoint(this._gunPosition, this._rotation, this._zoom);
        bulletPos.x = bulletPos.x + this._position.x;
        bulletPos.y = bulletPos.y + this._position.y;
        let bullet = new UfoBullet(this._mainDrawHelper, bulletPos);
        let dy = (userShipPosition.y - this._position.y);
        let dx = (userShipPosition.x - this._position.x);
        let fireAngleDegree = Utility.radiansToDegree(Math.atan(Math.abs(dy) / Math.abs(dx)));
        if (dx < 0 && dy > 0)
            fireAngleDegree = 180 - fireAngleDegree;
        else if (dx < 0 && dy < 0)
            fireAngleDegree += 180;
        else if (dx > 0 && dy < 0)
            fireAngleDegree = 360 - fireAngleDegree;
        bullet.vectorMotion.direction = fireAngleDegree;
        bullet.rotation = bullet.vectorMotion.direction;
        if (this._vectorMotion !== null) {
            if (bullet.vectorMotion.magnitude < this._vectorMotion.magnitude)
                bullet.vectorMotion.magnitude = this._vectorMotion.magnitude;
        }
        bullet.init();
        return bullet;
    }
}
class Spaceship extends GameObject {
    constructor(mainDrawHelper, position) {
        super(mainDrawHelper, position);
        this._zoom = 1.5;
    }
    buildObject() {
        let points = [new Point(-6, -4), new Point(0, 10), new Point(6, -4), new Point(5, -2), new Point(-5, -2)];
        this._polygons.add(points);
        this._gunPosition = new Point(0, 10);
    }
    accelerate() {
        this._vectorMotionChange = new Vector(AppParameters.accelerationIncrement, this._rotation);
    }
    stopAccelerate() {
        this._vectorMotionChange = null;
    }
    updatePosition() {
        if (this._vectorMotionChange === null) {
            if (this._vectorMotion !== null)
                this._vectorMotion.magnitude -= AppParameters.accelerationDecrement;
        }
        else {
            if (this._vectorMotion === null)
                this._vectorMotion = this._vectorMotionChange;
            else
                this._vectorMotion.add(this._vectorMotionChange);
        }
        if (this._vectorMotion !== null) {
            if (this._vectorMotion.magnitude === 0)
                this._vectorMotion = null;
            else if (this._vectorMotion.magnitude > AppParameters.maxSpeed)
                this._vectorMotion.magnitude = AppParameters.maxSpeed;
        }
        if (this._vectorMotion === null)
            return;
        let deltaX = this._vectorMotion.magnitude * Math.cos(this._vectorMotion.directionRad);
        let deltaY = this._vectorMotion.magnitude * Math.sin(this._vectorMotion.directionRad);
        this._position.x += deltaX;
        this._position.y += deltaY;
    }
    fire() {
        let bulletPos = Utility.transformPoint(this._gunPosition, this._rotation, this._zoom);
        bulletPos.x = bulletPos.x + this._position.x;
        bulletPos.y = bulletPos.y + this._position.y;
        let bullet = new Bullet(this._mainDrawHelper, bulletPos);
        bullet.vectorMotion.direction = this._rotation;
        bullet.rotation = this._rotation;
        if (this._vectorMotion !== null) {
            if (bullet.vectorMotion.magnitude < this._vectorMotion.magnitude)
                bullet.vectorMotion.magnitude = this._vectorMotion.magnitude;
        }
        bullet.init();
        return bullet;
    }
}
var GameObjectType;
(function (GameObjectType) {
    GameObjectType[GameObjectType["playerSpaceship"] = 0] = "playerSpaceship";
    GameObjectType[GameObjectType["bullet"] = 1] = "bullet";
    GameObjectType[GameObjectType["ufoShip"] = 2] = "ufoShip";
    GameObjectType[GameObjectType["ufoBullet"] = 3] = "ufoBullet";
    GameObjectType[GameObjectType["asteroid"] = 4] = "asteroid";
})(GameObjectType || (GameObjectType = {}));
var GameMode;
(function (GameMode) {
    GameMode[GameMode["startScreen"] = 0] = "startScreen";
    GameMode[GameMode["play"] = 1] = "play";
    GameMode[GameMode["gameOverScreen"] = 2] = "gameOverScreen";
})(GameMode || (GameMode = {}));
class Game {
    constructor() {
        this._gameMode = GameMode.startScreen;
        this._gameObjects = [];
        this._keyDown = new Map();
        this._keyUp = new Map();
        this._fireCount = 0;
        this._updateGameAreaCount = 0;
        this._ufoShips = [];
        this._playerLives = AppParameters.maxPlayerLives;
        this._playerScore = 0;
    }
    _clearGameArea() {
        this._context.beginPath();
        this._context.fillStyle = 'black';
        this._context.fillRect(0, 0, this._canvas.width, this._canvas.height);
    }
    _drawPlayerLives() {
        for (let i = 0; i < this._playerLives; i++) {
            let ship = new Spaceship(this._drawHelper, new Point(20 + i * 30, this._gameAreaHeight - 30));
            ship.zoom = 2;
            ship.init();
            ship.draw();
        }
    }
    _drawPlayerScore() {
        this._context.font = '24px Courier New';
        this._context.fillStyle = 'white';
        this._context.textAlign = "left";
        this._context.fillText(this._playerScore.toString(), 20, 75);
    }
    _manageUfoShips() {
        if (this._updateGameAreaCount % 1000 === 0) {
            // Add an Ufo ship when needed
            let ufoShip = new UfoShip(this._drawHelper, new Point(Utility.getRandomInteger(0, this._gameAreaWidth), Utility.getRandomInteger(0, this._gameAreaHeight)));
            ufoShip.init();
            if (this._ufoShips.length < AppParameters.maxUfoShips) {
                this._gameObjects.push(ufoShip);
                this._ufoShips.push(ufoShip);
            }
        }
        for (let ufoShip of this._ufoShips) {
            if (this._updateGameAreaCount % 100 === 0) {
                let bullet = ufoShip.fire(this._spaceShip.position);
                this._gameObjects.push(bullet);
            }
        }
    }
    _removeUfoShip(ufoShip) {
        let index = this._ufoShips.indexOf(ufoShip);
        if (index >= 0)
            this._ufoShips.splice(index, 1);
    }
    _splitOrDestroyAsteroid(asteroid) {
        if (asteroid.size > 1) {
            // Size > 1, decrease the size of the astoroid and create a new one of the same size
            asteroid.size--;
            let newAsteroid = new Asteroid(this._drawHelper, new Point(asteroid.position.x, asteroid.position.y));
            newAsteroid.size = asteroid.size;
            newAsteroid.type = asteroid.type;
            newAsteroid.rotation = asteroid.rotation;
            newAsteroid.vectorMotion.magnitude = asteroid.vectorMotion.magnitude;
            newAsteroid.vectorMotion.direction = asteroid.vectorMotion.direction - Utility.getRandomInteger(45, 270);
            newAsteroid.init();
            this._gameObjects.push(newAsteroid);
        }
        else {
            // Size = 1, destroy the asteroid
            asteroid.destroyed = true;
        }
    }
    _getObjectType(gameObject) {
        if (gameObject instanceof Spaceship)
            return GameObjectType.playerSpaceship;
        else if (gameObject instanceof UfoBullet)
            return GameObjectType.ufoBullet;
        else if (gameObject instanceof Bullet)
            return GameObjectType.bullet;
        else if (gameObject instanceof UfoShip)
            return GameObjectType.ufoShip;
        else
            return GameObjectType.asteroid;
    }
    _checkCollisions(gameObject) {
        for (let collisionObjIdx = 0; collisionObjIdx < this._gameObjects.length; collisionObjIdx++) {
            let collisionObject = this._gameObjects[collisionObjIdx];
            if (gameObject === collisionObject)
                continue;
            if (gameObject.destroyed)
                continue;
            let gameObjectType = this._getObjectType(gameObject);
            let collisionObjectType = this._getObjectType(collisionObject);
            if (!((gameObjectType === GameObjectType.bullet && collisionObjectType === GameObjectType.ufoShip)
                || (gameObjectType === GameObjectType.bullet && collisionObjectType === GameObjectType.asteroid)
                || (gameObjectType === GameObjectType.ufoBullet && collisionObjectType === GameObjectType.playerSpaceship)
                || (gameObjectType === GameObjectType.ufoBullet && collisionObjectType === GameObjectType.asteroid)
                || (gameObjectType === GameObjectType.playerSpaceship && collisionObjectType === GameObjectType.ufoShip)
                || (gameObjectType === GameObjectType.playerSpaceship && collisionObjectType === GameObjectType.asteroid)
                || (gameObjectType === GameObjectType.ufoShip && collisionObjectType === GameObjectType.asteroid)))
                continue;
            let bBox1 = gameObject.getBoundingBox();
            let bBox2 = collisionObject.getBoundingBox();
            if (AppParameters.debugDrawBoundingBox) {
                this._context.beginPath();
                this._context.strokeStyle = 'yellow';
                this._context.rect(bBox1.x, bBox1.y, bBox1.width, bBox1.height);
                this._context.rect(bBox2.x, bBox2.y, bBox2.width, bBox2.height);
                this._context.stroke();
            }
            if (!(bBox1.intersectsWith(bBox2) || bBox2.intersectsWith(bBox1)))
                continue;
            // The two objects intersects...
            switch (gameObjectType) {
                case GameObjectType.bullet:
                    if (collisionObjectType === GameObjectType.ufoShip) {
                        this._playerScore += AppParameters.scoreUfoShip;
                        collisionObject.destroyed = true;
                    }
                    else if (collisionObjectType === GameObjectType.asteroid) {
                        let asteroid = collisionObject;
                        switch (asteroid.size) {
                            case 1:
                                this._playerScore += AppParameters.scoreAsteroidSmall;
                                break;
                            case 2:
                                this._playerScore += AppParameters.scoreAsteroidMedium;
                                break;
                            case 3:
                                this._playerScore += AppParameters.scoreAsteroidBig;
                                break;
                        }
                        this._splitOrDestroyAsteroid(asteroid);
                    }
                    break;
                case GameObjectType.ufoBullet:
                    if (collisionObjectType === GameObjectType.playerSpaceship)
                        collisionObject.destroyed = true;
                    else if (collisionObjectType === GameObjectType.asteroid)
                        this._splitOrDestroyAsteroid(collisionObject);
                    break;
                case GameObjectType.playerSpaceship:
                    if (collisionObjectType === GameObjectType.ufoShip) {
                        this._playerScore += AppParameters.scoreUfoShip;
                        collisionObject.destroyed = true;
                    }
                    else if (collisionObjectType === GameObjectType.asteroid) {
                        let asteroid = collisionObject;
                        switch (asteroid.size) {
                            case 1:
                                this._playerScore += AppParameters.scoreAsteroidSmall;
                                break;
                            case 2:
                                this._playerScore += AppParameters.scoreAsteroidMedium;
                                break;
                            case 3:
                                this._playerScore += AppParameters.scoreAsteroidBig;
                                break;
                        }
                        this._splitOrDestroyAsteroid(asteroid);
                    }
                    break;
                case GameObjectType.ufoShip:
                    this._splitOrDestroyAsteroid(collisionObject);
                    break;
            }
            gameObject.destroyed = true;
            if (collisionObjectType === GameObjectType.playerSpaceship)
                this._playerSpaceshipHit(collisionObjIdx);
            if (collisionObject.destroyed) {
                this._gameObjects.splice(collisionObjIdx, 1);
                if (collisionObject instanceof UfoShip)
                    this._removeUfoShip(collisionObject);
            }
        }
    }
    _checkKeyDown(key) {
        if (this._keyDown.has(key))
            return this._keyDown.get(key);
        return false;
    }
    _resetKeyDown(key) {
        this._keyDown.set(key, false);
    }
    _checkKeyUp(key) {
        if (this._keyUp.has(key)) {
            if (this._keyUp.get(key)) {
                this._keyUp.set(key, false);
                return true;
            }
        }
        return false;
    }
    _manageUserInput() {
        if (this._checkKeyDown("ArrowUp")) {
            this._spaceShip.accelerate();
        }
        else {
            this._spaceShip.stopAccelerate();
        }
        if (this._checkKeyDown("ArrowLeft")) {
            this._spaceShip.rotation += AppParameters.rotationDecrement;
        }
        if (this._checkKeyDown("ArrowRight")) {
            this._spaceShip.rotation -= AppParameters.rotationIncrement;
        }
        if (this._checkKeyDown(" ")) {
            this._fireCount++;
            if (this._fireCount === 1 || this._fireCount % 10 === 0) {
                let bullet = this._spaceShip.fire();
                this._gameObjects.push(bullet);
            }
        }
        if (this._checkKeyUp(" ")) {
            this._fireCount = 0;
        }
    }
    _gameModeStartScreen() {
        this._clearGameArea();
        if (this._checkKeyDown("Enter")) {
            this._switchToPlayMode();
            return;
        }
        for (let currentObjIdx = 0; currentObjIdx < this._gameObjects.length; currentObjIdx++) {
            let gameObject = this._gameObjects[currentObjIdx];
            gameObject.updatePosition();
            gameObject.draw();
            if (gameObject.position.x > this._gameAreaWidth)
                gameObject.position.x = 0;
            else if (gameObject.position.x < 0)
                gameObject.position.x = this._gameAreaWidth;
            if (gameObject.position.y > this._gameAreaHeight)
                gameObject.position.y = 0;
            else if (gameObject.position.y < 0)
                gameObject.position.y = this._gameAreaHeight;
        }
        this._context.font = "100px Courier New";
        this._context.fillStyle = "white";
        this._context.textAlign = "center";
        this._context.fillText("Space Y", this._canvas.width / 2, this._canvas.height / 2);
        this._context.font = "25px Courier New";
        this._context.fillText("Press Enter to start", this._canvas.width / 2, (this._canvas.height / 2) + 60);
    }
    _playerSpaceshipHit(objectIndex) {
        this._spaceShip.destroyed = false;
        if (this._playerLives > 1) {
            this._playerLives--;
            this._spaceShip.position = new Point(this._gameAreaWidth / 2, this._gameAreaHeight / 2);
            this._spaceShip.rotation = 90;
            this._spaceShip.vectorMotion = null;
            return;
        }
        else {
            this._gameObjects.splice(objectIndex, 1);
            this._switchToGameOverScreenMode();
            return;
        }
    }
    _gameModePlay() {
        this._updateGameAreaCount++;
        this._manageUserInput();
        this._clearGameArea();
        this._drawPlayerLives();
        this._manageUfoShips();
        for (let currentObjIdx = 0; currentObjIdx < this._gameObjects.length; currentObjIdx++) {
            let gameObject = this._gameObjects[currentObjIdx];
            // In case of bullets, spaceship and ufoships, check for collisions with other objects
            if (gameObject instanceof Bullet || gameObject instanceof Spaceship || gameObject instanceof UfoShip)
                this._checkCollisions(gameObject);
            // Check if the object has been destroyed (after collision detection code)
            if (this._spaceShip.destroyed) {
                this._playerSpaceshipHit(currentObjIdx);
            }
            if (gameObject.destroyed) {
                this._gameObjects.splice(currentObjIdx, 1);
                if (gameObject instanceof UfoShip)
                    this._removeUfoShip(gameObject);
            }
            else {
                // Update object position and check if outside the game area
                // If so bring it back inside!
                gameObject.updatePosition();
                if (gameObject.position.x > this._gameAreaWidth)
                    gameObject.position.x = 0;
                else if (gameObject.position.x < 0)
                    gameObject.position.x = this._gameAreaWidth;
                if (gameObject.position.y > this._gameAreaHeight)
                    gameObject.position.y = 0;
                else if (gameObject.position.y < 0)
                    gameObject.position.y = this._gameAreaHeight;
                // Draw the object
                gameObject.draw();
            }
            this._drawPlayerScore();
            if (this._gameObjects.filter(t => t instanceof Asteroid).length === 0) {
                // No more asteroids, level completed
                this._switchToPlayMode();
                return;
            }
        }
    }
    _gameModeGameOver() {
        this._clearGameArea();
        if (this._checkKeyDown("Enter")) {
            this._resetKeyDown("Enter");
            this._switchToStartScreenMode();
            return;
        }
        for (let currentObjIdx = 0; currentObjIdx < this._gameObjects.length; currentObjIdx++) {
            let gameObject = this._gameObjects[currentObjIdx];
            gameObject.updatePosition();
            gameObject.draw();
            if (gameObject.position.x > this._gameAreaWidth)
                gameObject.position.x = 0;
            else if (gameObject.position.x < 0)
                gameObject.position.x = this._gameAreaWidth;
            if (gameObject.position.y > this._gameAreaHeight)
                gameObject.position.y = 0;
            else if (gameObject.position.y < 0)
                gameObject.position.y = this._gameAreaHeight;
        }
        this._context.font = "100px Courier New";
        this._context.fillStyle = "white";
        this._context.textAlign = "center";
        this._context.fillText("Game Over", this._canvas.width / 2, this._canvas.height / 2);
        this._context.font = "25px Courier New";
        this._context.fillText("Press Enter to continue", this._canvas.width / 2, (this._canvas.height / 2) + 60);
    }
    _updateGameArea() {
        if (this._gameMode === GameMode.startScreen)
            this._gameModeStartScreen();
        else if (this._gameMode === GameMode.gameOverScreen)
            this._gameModeGameOver();
        else
            this._gameModePlay();
    }
    _onKeyDown(e) {
        this._keyDown.set(e.key, true);
        this._keyUp.set(e.key, false);
        e.stopImmediatePropagation();
        e.preventDefault();
    }
    _onKeyUp(e) {
        this._keyDown.set(e.key, false);
        this._keyUp.set(e.key, true);
        e.stopImmediatePropagation();
        e.preventDefault();
    }
    _resizeWindow(e) {
        this._gameAreaWidth = this._canvas.width;
        this._gameAreaHeight = this._canvas.height;
    }
    init() {
        this._body = document.getElementsByTagName("body")[0];
        this._canvas = document.getElementById("GameArea");
        this._context = this._canvas.getContext("2d");
        this._body.addEventListener("keydown", (e) => this._onKeyDown(e));
        this._body.addEventListener("keyup", (e) => this._onKeyUp(e));
        window.addEventListener("resize", (e) => this._resizeWindow(e));
        this._resizeWindow(null);
        this._axisOrigin = new Point(0, this._gameAreaHeight);
        this._drawHelper = new DrawHelper(this._context, this._axisOrigin);
        this._switchToStartScreenMode();
        setInterval(() => { this._updateGameArea(); }, AppParameters.gameAreaUpdateDelayMS);
    }
    _switchToStartScreenMode() {
        this._gameMode = GameMode.startScreen;
        this._playerLives = AppParameters.maxPlayerLives;
        this._playerScore = 0;
        this._gameObjects = [];
        this._ufoShips = [];
        for (let i = 0; i < AppParameters.maxAsteroids * 3; i++) {
            let asteroid = new Asteroid(this._drawHelper, new Point(Utility.getRandomInteger(0, this._gameAreaWidth), Utility.getRandomInteger(0, this._gameAreaHeight)));
            asteroid.init();
            this._gameObjects.push(asteroid);
        }
    }
    _switchToPlayMode() {
        this._gameMode = GameMode.play;
        this._gameObjects = [];
        this._ufoShips = [];
        this._spaceShip = new Spaceship(this._drawHelper, new Point(this._gameAreaWidth / 2, this._gameAreaHeight / 2));
        this._gameObjects.push(this._spaceShip);
        for (let i = 0; i < AppParameters.maxAsteroids; i++) {
            this._gameObjects.push(new Asteroid(this._drawHelper, new Point(Utility.getRandomInteger(0, this._gameAreaWidth), Utility.getRandomInteger(0, this._gameAreaHeight))));
        }
        for (let g of this._gameObjects) {
            g.init();
        }
    }
    _switchToGameOverScreenMode() {
        this._gameMode = GameMode.gameOverScreen;
    }
}
let GI = new Game();
GI.init();
let GameArea = document.getElementById("GameArea");
//# sourceMappingURL=game.js.map